#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import time
import sys
import os
import datetime
import logging
import threading
import platform
import requests
import random

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

from waptutils import __version__

import socketio

import logging
import sqlite3

import json
import io

import _thread
import threading

import queue
import traceback
import locale

import datetime
import copy

import tempfile

import setuphelpers

# wapt specific stuff
from waptutils import ensure_unicode, ensure_list, jsondump, is_between_two_times
from waptutils import get_requests_client_cert_session,default_http_headers
from waptutils import harakiri
from waptpackage import PackageEntry, WaptLocalRepo, WaptPackage, EWaptException
from waptcrypto import SSLVerifyException, SSLCABundle, SSLCertificate, SSLPrivateKey
from urllib3.exceptions import InsecureRequestWarning
from requests.exceptions import HTTPError

from common import Wapt

from waptservice.waptservice_common import waptservice_remote_actions, waptconfig, WaptServiceConfig
from waptservice.waptservice_common import WaptUpdate, WaptUpgrade, WaptUpdateServerStatus, WaptRegisterComputer
from waptservice.waptservice_common import WaptCleanup, WaptPackageInstall, WaptPackageRemove, WaptPackageForget, WaptLongTask, WaptAuditPackage

from waptservice.repositories import WaptSyncRepo
from waptservice.enterprise import WaptRunSessionSetup


if sys.platform == 'win32':
    import pythoncom

from waptservice.plugins import *

logger = logging.getLogger('waptws')
tasks_logger = logging.getLogger('wapttasks')

# Websocket stuff
##### API V2 #####


def make_response(result={}, success=True, error_code='', msg='', uuid=None, request_time=None):
    data = dict(
        uuid=uuid,
        success=success,
        msg=msg,
        error_code=error_code,
        result=result,
        request_time=request_time,
    )
    return data


def make_response_from_exception(exception, error_code='', uuid=None, request_time=None):
    """Create a standard answer for websocket callback exception

    Returns:
        dict: {success : False, msg : message from exception, error_code : classname of exception if not provided}
   """
    if not error_code:
        error_code = type(exception).__name__.lower()

    data = dict(
        uuid=uuid,
        success=False,
        msg="Error on client: %s" % repr(exception),
        error_code=error_code,
        request_time=request_time,
    )
    return data


def wait_for_event_send_tasks(task_manager, last_received_event_id, timeout):
    data = None
    start_time = time.time()
    while True:
        actual_last_event_id = task_manager.events.last_event_id()
        if actual_last_event_id is not None and actual_last_event_id <= last_received_event_id:
            data = {'last_event_id': task_manager.events.last_event_id()}
            if time.time() - start_time > timeout:
                break
        elif actual_last_event_id is None or actual_last_event_id > last_received_event_id:
            data = task_manager.tasks_status()
            break

        if time.time() - start_time > timeout:
            break

        # avoid eating cpu
        time.sleep(0.1)

    # send back the tasks to the console
    return data

class WaptSocketIORemoteCalls(socketio.ClientNamespace):
    _wapt = None

    @property
    def wapt(self):
        # thread local wapt...
        if not hasattr(self,'_wapt') or self._wapt is None:
            self._wapt = Wapt(config_filename = self.config_filename)
            tasks_logger.info('Socketio new Wapt intance has Thread ID: %s' % threading.current_thread().ident)
        return self._wapt

    def on_trigger_host_action(self, args):
        try:
            start_time = time.time()
            actions = args
            if not isinstance(actions, list):
                actions = [actions]
            # check signatures
            if not self.wapt:
                raise Exception('Wapt not available')
            verified_by = None
            for action in actions:
                tasks_logger.info('Host actions "%s" triggered by SocketIO by %s' % (action['action'], action['signer']))
                name = action['action']
                verified_by = None
                # full cert chain provided with the signed action
                if action.get('signer_certificate', None):
                    signer_cert_ca = SSLCABundle()
                    signer_cert_ca.add_certificates_from_pem(action['signer_certificate'].encode('utf8'))
                    signer_cert_chain = signer_cert_ca.certificates()
                else:
                    # only sha256 fingerprint provided. (lighter). issuer must be in the authorized cabundle
                    signer_cert = self.wapt.cabundle.certificate(action['signer'])
                    if signer_cert:
                        signer_cert_chain = [signer_cert]
                    else:
                        signer_cert_chain = []
                chain = self.wapt.cabundle.check_certificates_chain(signer_cert_chain)
                if chain:
                    required_attributes = ['uuid', 'action']
                    if name in ['trigger_install_packages', 'trigger_remove_packages', 'trigger_forget_packages']:
                        required_attributes.append('packages')
                    if name in ['trigger_change_description']:
                        required_attributes.append('computer_description')
                    if name in waptservice_remote_actions:
                        required_attributes.extend(waptservice_remote_actions[name].required_attributes)

                    verified_by = chain[0].verify_claim(action, max_age_secs=waptconfig.signature_clockskew,
                                                        required_attributes=required_attributes)
                else:
                    raise SSLVerifyException('Untrusted certificate %s for signed action %s, aborting' % (signer_cert_chain, action))

                if not verified_by:
                    raise SSLVerifyException('Bad signature for action %s, aborting' % action)

            if verified_by:
                verified_by = verified_by.get('verified_by', None)

            result = []
            for action in actions:
                uuid = action['uuid']
                if uuid != self.wapt.host_uuid:
                    raise Exception('Task is not targeted to this host. task''s uuid does not match host''uuid')
                name = action['action']
                if name in ['trigger_cancel_all_tasks']:
                    data = [t.as_dict() for t in self.task_manager.cancel_all_tasks()]
                    result.append(data)

                elif name in ['trigger_host_update', 'trigger_host_register', 'trigger_host_update_server_status']:
                    if name == 'trigger_host_update_server_status':
                        task = WaptUpdateServerStatus(force=True,include_wmi=True,include_dmi=True)
                    elif name == 'trigger_host_update':
                        task = WaptUpdate()
                        task.force = action.get('force', False)
                    elif name == 'trigger_host_register':
                        task = WaptRegisterComputer(computer_description=action.get('computer_description', None))
                        task.force = action.get('force', False)
                    task.notify_user = action.get('notify_user', False)
                    task.notify_server_on_finish = action.get('notify_server', False)
                    task.created_by = verified_by
                    tasks_logger.info('task "%s" to %s triggered by %s' % (task,name,verified_by))
                    data = self.task_manager.add_task(task).as_dict()
                    result.append(data)

                elif name in ['trigger_change_description']:
                    desc = action.get('computer_description', None)
                    tasks_logger.info('task "Change host description" to %s triggered by %s' % (desc,verified_by))
                    if desc is not None:
                        setuphelpers.set_computer_description(desc)
                        msg = 'Computer description of %s changed to %s' % (setuphelpers.get_hostname(), setuphelpers.get_computer_description())
                        if not setuphelpers.get_computer_description() == desc:
                            raise Exception('Computer description has not been changed')
                        result.append(dict(success=True,
                                           msg=msg,
                                           result=msg,
                                           ))
                        if action.get('notify_server', False):
                            task = WaptUpdate(created_by=verified_by)
                            task.notify_server_on_finish = True
                            self.task_manager.add_task(task)
                            result.append(task.as_dict())

                elif name == 'trigger_host_upgrade':
                    tasks_logger.info('task "Host upgrade" triggered remotely by %s' % (verified_by))
                    notify_user = action.get('notify_user', False)
                    notify_server_on_finish = action.get('notify_server', False)
                    force = action.get('force', False)
                    only_priorities = action.get('only_priorities', None)
                    only_if_not_process_running = action.get('only_if_not_process_running', False)
                    update_packages = action.get('update', True)

                    if update_packages:
                        result.append(self.task_manager.add_task(WaptUpdate(force=force, notify_user=notify_user,
                                                                            )).as_dict())

                    result.append(self.task_manager.add_task(WaptUpgrade(notify_user=notify_user,
                                                                         created_by=verified_by,
                                                                         only_priorities=only_priorities,
                                                                         only_if_not_process_running=only_if_not_process_running,
                                                                         force=force
                                                                         )).as_dict())

                    result.append(self.task_manager.add_task(WaptCleanup(notify_user=False, created_by=verified_by, priority=200)).as_dict())
                    if self.wapt.is_enterprise() and setuphelpers.get_loggedinusers():
                        result.append(self.task_manager.add_task(WaptRunSessionSetup()).as_dict())

                elif name in ['trigger_install_packages', 'trigger_remove_packages', 'trigger_forget_packages']:
                    packagenames = action['packages']
                    only_priorities = action.get('only_priorities', None)
                    only_if_not_process_running = action.get('only_if_not_process_running', False)

                    for packagename in packagenames:
                        if name == 'trigger_install_packages':
                            task = WaptPackageInstall(packagenames=packagename)
                        elif name == 'trigger_remove_packages':
                            task = WaptPackageRemove(packagenames=packagename)
                        elif name == 'trigger_forget_packages':
                            task = WaptPackageForget(packagenames=packagename)
                        task.force = action.get('force', False)
                        task.notify_user = action.get('notify_user', False)
                        task.notify_server_on_finish = action.get('notify_server', False)
                        task.created_by = verified_by
                        task.only_priorities = only_priorities
                        task.only_if_not_process_running = only_if_not_process_running

                        tasks_logger.info('Task %s triggered remotely by %s' % (task,verified_by))
                        result.append(self.task_manager.add_task(task).as_dict())

                    if name == 'trigger_install_packages':
                        self.task_manager.add_task(WaptAuditPackage(packagenames=packagenames,
                                                                    created_by=verified_by,
                                                                    force=task.force,
                                                                    notify_user=task.notify_user,
                                                                    notify_server_on_finish=task.notify_server_on_finish,
                                                                    priority=200)).as_dict()
                        if self.wapt.is_enterprise():
                            if setuphelpers.get_loggedinusers():
                                self.task_manager.add_task(WaptRunSessionSetup())


                elif name == 'trigger_waptservicerestart':
                    tasks_logger.info('Waptservice restart triggered by %s' % verified_by)
                    try:
                        if platform.system() == 'Windows':
                            harakiri(10)

                        elif platform.system() == 'Darwin':
                            msg = setuphelpers.run('launchctl unload /Library/LaunchDaemons/com.tranquilit.tis-waptagent.plist;launchctl trigload /Library/LaunchDaemons/com.tranquilit.tis-waptagent.plist;')
                        else:
                            msg = setuphelpers.run('systemctl restart waptservice')
                        result.append(dict(success=True, msg=msg, result=msg))
                    except:
                        # restart by nssm
                        harakiri(10)
                elif name == 'unregister_computer':
                    self.wapt.delete_param('last_update_server_hashes')
                    self.wapt.delete_param('last_audit_data_server_date')
                    self.wapt.delete_param('server_uuid')
                    if os.path.isfile(self.wapt.get_host_certificate_filename()):
                        os.unlink(self.wapt.get_host_certificate_filename())
                    self.disconnect()

                elif name == 'trigger_longtask':
                    task = WaptLongTask()
                    task.force = args.get('force', False)
                    task.notify_user = args.get('notify_user', False)
                    task.notify_server_on_finish = args.get('notify_server', False)
                    task.created_by = verified_by
                    result.append(self.task_manager.add_task(task).as_dict())
                elif name in waptservice_remote_actions:
                    waptservice_remote_actions[name].trigger_action(self, action, verified_by)
                else:
                    raise EWaptException('Unhandled remote action %s' % name)

            # self.emit('trigger_update_result',{'result':data})
            #if result_callback:
            return make_response(result, uuid=self.wapt.host_uuid, request_time=time.time()-start_time)
        except BaseException as e:
            logger.warning('Exception for actions %s: %s' % (repr(args), repr(e)))
            #if result_callback:
            #    result_callback(make_response_from_exception(e, uuid=self.wapt.host_uuid, request_time=time.time()-start_time))

    def on_get_tasks_status(self, args):
        # check signatures
        try:
            if not self.wapt:
                raise Exception('Wapt not available')

            uuid = args.get('uuid', '')
            if uuid != self.wapt.host_uuid:
                raise Exception('Task is not targeted to this host. task''s uuid does not match host''uuid')

            timeout = float(args.get('timeout', '10.0'))
            last_received_event_id = int(args.get('last_event_id', '-1'))

            data = {}
            if self.task_manager.events:
                data = wait_for_event_send_tasks(self.task_manager, last_received_event_id, timeout)
            else:
                time.sleep(0.5)

            return make_response(result = data, uuid=self.wapt.host_uuid)

        except BaseException as e:
            logger.info('Exception for actions %s: %s' % (repr(args), repr(e)))
            return make_response_from_exception(e, uuid=self.wapt.host_uuid)

    def on_wapt_ping(self, args=None):
        logger.debug('wapt_ping... %s' % (args,))
        self.emit('wapt_pong')

    def on_sync_remote_repo(self, args=None):
        tasks_logger.info('Synchronize local remote repo %s' % (args,))
        if waptconfig.enable_remote_repo and self.wapt.is_enterprise() and not(waptconfig.sync_only_forced):
            if not(waptconfig.local_repo_time_for_sync_start) or is_between_two_times(waptconfig.local_repo_time_for_sync_start, waptconfig.local_repo_time_for_sync_end):
                try:
                    self.task_manager.add_task(WaptSyncRepo(notifyuser=False, created_by='SERVER', socketio_client=self))
                except Exception as e:
                    logger.debug('Error syncing local repo with server repo : %s' % e)
            else:
                self.emit("synchronization_not_in_time_range")
        else:
            self.emit("synchronization_not_a_local_remote_repo")

    def on_sync_remote_repo_force(self, args=None):
        tasks_logger.info('Synchronize local remote repo %s' % (args,))
        if waptconfig.enable_remote_repo and self.wapt.is_enterprise():
            try:
                self.task_manager.add_task(WaptSyncRepo(notifyuser=False, created_by='SERVER', socketio_client=self))
            except Exception as e:
                logger.debug('Error syncing local repo with server repo : %s' % e)
        else:
            self.emit("synchronization_not_a_local_remote_repo")

    def on_audit_remote_repo(self, args=None):
        tasks_logger.info('Auditing files of remote repo %s' % (args,))
        if waptconfig.enable_remote_repo and self.wapt.is_enterprise():
            try:
                self.task_manager.add_task(WaptSyncRepo(notifyuser=False, created_by='SERVER', audit=True, socketio_client=self))
            except Exception as e:
                logger.debug('Error auditing local repo with server repo : %s' % e)
        else:
            self.emit("synchronization_not_a_local_remote_repo")

    def on_wapt_force_reconnect(self, args=None):
        tasks_logger.info('Force disconnect from server...')
        self.disconnect()

    def on_message(self, message=None):
        logger.debug('socket.io message : %s' % message)

    def on_event(self, event, *args):
        logger.debug('socket.io event : %s, args: %s' % (event, args))


def get_requests_socketclient_session(url=None, cert=None, verify=True, proxies={'http': None, 'https': None}, **kwargs):
    """Returns a requests Session which is aware of client cert auth with password protected key
    Disable use of environ.

    Args:
        url (str): base prefix url for which the session is created
        cert (tuple) : (certfilename,pem encoded key filename, key password)
        verify (bool or str) : verify server certificate. Id str, path to trusted CA bundle

    Returns:
        Session
    """
    result = requests.Session()
    # be sure to not use HTTP_PROXY or HTTPS_PROXY environ variable
    result.trust_env = False
    result.headers = default_http_headers()
    result.verify = verify
    result.proxies = proxies
    if not verify:
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)  # pylint: disable=no-member

    result.cert = cert
    return result


class WaptSocketIOClient(threading.Thread):
    def __init__(self, config_filename='c:/wapt/wapt-get.ini', task_manager=None, request_timeout=None):
        threading.Thread.__init__(self)
        self.name = 'SocketIOClient'
        self.config_filename = config_filename
        self.task_manager = task_manager
        self.config = WaptServiceConfig(config_filename)
        self.socketio_client = None
        self.server_authorization_token = None
        self.connect_params = {}
        if request_timeout is None:
            request_timeout = self.config.websockets_request_timeout
        self.request_timeout = request_timeout

    def websocket_url(self):
        return "%s://%s:%s" % (self.config.websockets_proto, self.config.websockets_host,self.config.websockets_port)

    def run(self):
        if sys.platform == 'win32':
            pythoncom.CoInitializeEx(pythoncom.COINIT_MULTITHREADED)

        self.config.load()
        with Wapt(config_filename=self.config.config_filename) as tmp_wapt:
            tasks_logger.info('Starting socketio on "%s://%s:%s" Thread id %s' % (self.config.websockets_proto, self.config.websockets_host, self.config.websockets_port, threading.current_thread().ident))
            logger.debug('Certificate checking : %s' % self.config.websockets_verify_cert)

            def update_connect_params(wapt):
                config_updated = wapt.reload_config_if_updated()
                if config_updated:
                    self.config.load()
                if (self.socketio_client and not self.socketio_client.connected) or config_updated or not self.server_authorization_token or not self.connect_params or self.connect_params.get('uuid') != wapt.host_uuid:
                    try:
                        self.server_authorization_token = wapt.get_auth_token('websocket')
                        logger.info('Websocket token: %s' % self.server_authorization_token)
                        self.connect_params['uuid'] = wapt.host_uuid
                        self.connect_params['Authorization'] = 'Bearer %s' % self.server_authorization_token
                        return True
                    except Exception as e:
                        self.server_authorization_token = None
                        self.connect_params['uuid'] = wapt.host_uuid
                        if 'Authorization' in self.connect_params:
                            del self.connect_params['Authorization']
                        if isinstance(e,EWaptException) or (isinstance(e,HTTPError) and e.response.status_code in (400,401)):
                            logger.warning('Websocket connect params failure, trying registration: %s' % e)
                            try:
                                register_result = wapt.register_computer()
                                wapt.reload_config_if_updated()
                                tasks_logger.info('New registration successful')
                                wapt.get_auth_token('websocket')
                                logger.info('Websocket token: %s' % self.server_authorization_token)
                                self.connect_params['uuid'] = wapt.host_uuid
                                self.connect_params['Authorization'] = 'Bearer %s' % self.server_authorization_token
                                return True
                            except Exception as register_exception:
                                tasks_logger.warning('Unable to register: %s'% register_exception)
                                raise
                        else:
                            raise

                    except Exception as e:
                        self.server_authorization_token = None
                        self.connect_params['uuid'] = wapt.host_uuid
                        if 'Authorization' in self.connect_params:
                            del self.connect_params['Authorization']
                        raise
                return False

            while True:
                try:
                    need_reconnect = update_connect_params(tmp_wapt)
                    if need_reconnect:
                        tasks_logger.info('Socketio connection params have changed. Socketio needs reconnect')

                    if not self.socketio_client and self.config.websockets_host:
                        if self.config.waptserver.proxies and self.config.waptserver.proxies.get(self.config.websockets_proto, None) is not None:
                            proxies = self.config.waptserver.proxies
                        else:
                            proxies = None

                        cert = (tmp_wapt.get_host_certificate_filename(), tmp_wapt.get_host_key_filename())

                        tasks_logger.info('Creating socketio client: %s client auth cert: %s proxies: %s verify_cert: %s' % (self.websocket_url(),cert,proxies,self.config.websockets_verify_cert))

                        self.socketio_client = socketio.Client(
                            reconnection = True,
                            reconnection_attempts = 5,
                            reconnection_delay_max = self.config.websockets_retry_delay - 2,
                            logger = logger,
                            engineio_logger = logger,
                            request_timeout=self.request_timeout,

                            ssl_verify=self.config.websockets_verify_cert,
                            http_session = get_requests_socketclient_session(url=self.websocket_url(),
                                cert=cert,
                                proxies=proxies,
                                verify=self.config.websockets_verify_cert
                                ),
                            )

                    if self.socketio_client and self.config.websockets_host:
                        if self.socketio_client.connected and need_reconnect:
                            tasks_logger.info('Disconnecting Socketio')
                            self.socketio_client.disconnect()
                            self.socketio_client.wait()

                            #start = time.time()
                            #while (self.socketio_client.eio.read_loop_task.isAlive() or self.socketio_client.eio.write_loop_task.isAlive()) and (time.time() - start < 10):
                            #    time.sleep(1)

                        if not self.socketio_client.connected:
                            if '/' in self.socketio_client.namespace_handlers:
                                name_space = self.socketio_client.namespace_handlers['/']
                            else:
                                name_space = WaptSocketIORemoteCalls()
                                self.socketio_client.register_namespace(name_space)
                            name_space.config_filename = self.config_filename
                            name_space.task_manager = self.task_manager
                            name_space.config = self.config

                            tasks_logger.info('Connecting Socketio to %s' % (self.websocket_url(),))
                            self.socketio_client.connect(
                                url=self.websocket_url(),
                                headers = self.connect_params,
                                socketio_path = self.config.websockets_root,
                                transports = ['websocket','polling'])

                        if self.socketio_client.connected:
                            tasks_logger.info('WS read loop for %ss' % (self.config.websockets_check_config_interval,))
                            self.socketio_client.sleep(1)  # give the reconnect task time to start up
                            if self.socketio_client.eio.read_loop_task:
                                # read messages for websockets_check_config_interval seconds.
                                # when elapsed, we will check config.
                                self.socketio_client.eio.read_loop_task.join(self.config.websockets_check_config_interval)
                            else:
                                # we should retry later.
                                tasks_logger.warning('WS EIO read loop failed')
                                self.socketio_client.sleep(10)

                    else:
                        time.sleep(self.config.websockets_retry_delay)

                except Exception as e:
                    sleep_sec = self.config.websockets_retry_delay + random.randint(0,20)
                    logger.warning('Exception %s, waiting %ss before retrying' %
                                (repr(e), sleep_sec))
                    time.sleep(sleep_sec)


if __name__ == "__main__":
    pass
